#!/usr/bin/ruby
#
# serverstats 0.1
#
# Author::    Henning Brinkmann  (mailto:serverstats@henning-brinkmann.de)
# Copyright:: Copyright (c) 2011 Henning Brinkmann
# License::   GNU Public License
#
#
# serverstats is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# serverstats is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "socket"

# This module provides classes for graphs representing server
# statistics.
module ServerstatGraphs
  require "./zfs"
  require 'digest/md5'

  # This class is the base class for all graphs provided by this
  # module.
  class Graph
    # Initializer
    #
    # Sets the width attribute to 512. The y-axis is scaled linear by
    # default.
    #
    # [name] the name of the graph
    def initialize(name)
      @name = name
      @width = 512
      @logarithmic = false
    end

    # Set the width of the graph.
    #
    # [width] the width
    def width=(width)
      @width = width
    end

    # Set wether the y-axis shall be scaled logarithmically.
    #
    # [logarithmic] true: scale logarithmically, false: scale linear
    def logarithmic=(logarithmic)
      @logarithmic = logarithmic
    end
    
    # Returns the file name for the graph.
    #
    # The name of the file is generated like this:
    #
    # <tt> <GRAPHPATH>/@name.png </tt>
    def filename
      return ENV['GRAPHPATH'] + "/" + @name + ".png"
    end

    # Returns the name of the graph.
    def name()
      @name
    end
    
    # Returns the rrdgraph definitions for the graph.
    def defs
      return Array.new
    end

    # Returns the rrdgraph definitions for the graph.
    def graphs
      return Array.new
    end

    # Returns the rrdgraph command for the graph.
    def rrdgraph
      command = "rrdtool graph #{filename} -a PNG"
      if @width != nil then
        command += " -w #{@width}"
      end
      if @logarithmic then
        command += " -o"
      end
      command += " \\\n    " + defs.join(" \\\n    ");
      command += " \\\n    " + graphs.join(" \\\n    ");
      
      return command
    end
  end
  
  # A simple graph. 
  #
  # Represents a single graph.
  class SimpleGraph < Graph
    # Initializer
    #
    # Note: @stacked does only take effect in GroupedGraph.
    #
    # [name] the name of the graph
    # [file] RRD file containing the data
    # [ds] RRD dataset to display
    # [cf] RRD consolidation function
    # [kind] RRD kind of graph (AREA, LINE, ...)
    # [color] color of the graph
    # [stacked] stack graph if set to true
    def initialize(name, file, ds, cf, kind, color, stacked = false)
      super name
      @file = file
      @ds = ds
      @cf = cf
      @kind = kind
      @color = color
      @stacked = stacked
    end

    # Returns the rrdgraph definitions for the graph.
    #
    # Generates a string following this scheme:
    #
    # <tt>DEF:#{@name}=#{@file}:#{@ds}:#{@cf}</tt>
    #
    # <tt>VDEF:v_#{@name}=#{@name},AVERAGE</tt>
    def defs
      result = Array.new
      result << "DEF:#{@name}=#{@file}:#{@ds}:#{@cf}"
      result << "VDEF:v_#{@name}=#{@name},AVERAGE"
      
      return result
    end

    # Returns the rrdgraph graphs for the graph.
    #
    # Generates a string following this scheme:
    #
    # <tt>#{@kind}:#{@name}##{@color}:#{@name}</tt>
    def graphs
      result = Array.new
      graph = "#{@kind}:#{@name}##{@color}:#{@name}"

      if @stacked then
        graph << ":STACK"
      end
      
      result << graph

      result << "GPRINT:v_#{@name}:\"%3.3lf %s\\l\""

      return result
    end
  end

  # A grouped graph
  #
  # Groups several instances of Graph.
  class GroupedGraph < Graph
    # Initializer
    #
    # [name] name of the graph
    def initialize(name)
      super name
      @group = Array.new
    end

    # Adds a graph.
    #
    # [graph] instance of Graph to add
    def add(graph)
      @group << graph
    end

    # Returns rrdgraph definitions for this graph.
    #
    # Returns array containing all definitions of the graphs added by
    # add.
    def defs
      result = @group.collect do |graph|
        graph.defs
      end
      result = result.flatten

      return result
    end

    # Returns rrdgraph graph definitions for this graph.
    #
    # Returns array containing all graph definitions of graphs added
    # by add.
    def graphs
      result = @group.collect do |graph|
        graph.graphs
      end
      result = result.flatten

      return result
    end
  end

  # A graph representing read and written bytes of ZFS pool.
  class ZPoolGraph < Graph
    # Initializer
    #
    # [name] name of the graph
    # [pool] name of the ZFS pool
    def initialize(name, pool)
      super name
      @pool = pool
      @info = ZFS::ZpoolInfo.new(@pool)
    end

    # Generates an rpn expression.
    #
    # [first] first operand
    # [second] second operand
    # [op] operator
    #
    # Returns 
    # [first] if <i>second</i> is nil
    # [second] if <i>first</i> is nil
    # [first, second, op] otherwise
    def rpn(first, second, op)
      if first == nil
        return second
      elsif second == nil
        return first
      end

      return "#{first},#{second},#{op}"
    end

    # Generates a rrdgraph DEF.
    #
    # [name] the name of the DEF
    # [sd] the disk identifier
    # [ds] the dataset
    # [cf] the consolidation function
    #
    # The resulting string obeys the following scheme:
    #
    # <tt>DEF:#{name}=#{file}:#{ds}:#{cf}</tt>
    #
    # <tt>file</tt> is determined as the first file RRDPATH that
    # satisfies the pattern
    #
    # <tt>zpool_#{@pool}*#{sd}_#{ds}.rrd</tt>
    def gendef(name, sd, ds, cf)
      file = Dir["zpool_#{@pool}*#{sd}_#{ds}.rrd"][0]
      return "DEF:#{name}=#{file}:#{ds}:#{cf}"
    end

    # Returns rrdgraph defnitions for the graph.
    def defs
      sds = @info.getSDs

      result = Array.new
      writtenStacked = nil
      readStacked = "0"
      writtenMaxStacked = nil
      readMaxStacked = "0"
      sds.each do |sd| # for each disk
        # written bytes for current sd
        sdwritten = "#{sd}_nwritten"
        result << gendef(sdwritten, sd, "nwritten", "AVERAGE")

        # The stacked graph for the current sd is the sum of the
        # previous sd's graphs, plus the single graph of the current
        # sd.
        writtenStacked = rpn(writtenStacked, sdwritten, "+")
          
        sdwrittenstacked = "#{sdwritten}_stacked"
        result << "CDEF:#{sdwrittenstacked}=#{writtenStacked}"

        sdwrittenmax = "#{sd}_nwritten_max"
        result << gendef(sdwrittenmax, sd, "nwritten", "MAX")
        
        # See comment for writtenStacked. Analogous for maximum of
        # written bytes.
        writtenMaxStacked = rpn(writtenMaxStacked, sdwrittenmax, "+")

        # read bytes for current sd
        sdread = "#{sd}_nread"
        result << gendef(sdread, sd, "nread", "AVERAGE")

        # See comment for writtenStacked. The difference here is the
        # "-". Read bytes are shown downwards.
        readStacked = rpn(readStacked, sdread, "-")
          
        sdreadstacked = "#{sdread}_stacked"
        result << "CDEF:#{sdreadstacked}=#{readStacked}"

        sdreadmax = "#{sd}_nread_max"
        result << gendef(sdreadmax, sd, "nread", "MAX")

        # See comment for writtenStacked. Analogous for maximum of
        # read bytes. Again, read bytes are shown downwards. Hence
        # "-".
        readMaxStacked = rpn(readMaxStacked, sdreadmax, "-")
      end

      result << "CDEF:written_max=#{writtenMaxStacked}"
      result << "CDEF:read_max=#{readMaxStacked}"
        
      return result
    end

    # Returns the rrdgraph graphs for the graph.
    def graphs
      sds = @info.getSDs
      result = Array.new

      result << "AREA:written_max#777777"
      result << "AREA:read_max#777777"

      index = sds.count
      sds.reverse.each do |sd|
        val = 0x77 + (0xff - 0x77) / sds.count * index
        colorRead = sprintf("%06x", val * 0x00ff00 / 0xff)
        colorWritten = sprintf("%06x", val * 0xff0000 / 0xff)
        result << 
          "AREA:#{sd}_nwritten_stacked##{colorWritten}:\"#{sd} written\""
        result << 
          "AREA:#{sd}_nread_stacked##{colorRead}:\"#{sd} read\""
        
        index -= 1
      end
      
      result << "HRULE:0#000000"

      return result
    end
  end

  # A graph showing traffic on a network device.
  class NetworkGraph < Graph
    # Initializer
    #
    # [name] the name of the graph
    # [device] the name of the device
    def initialize(name, device)
      super name
      @device = device
    end

    # Returns the rrdgraph definitions for this graph.
    def defs
      prefix ="kstat_#{@device}"
      result = Array.new
      result << "DEF:obytes=" + Dir["#{prefix}*obytes.rrd"][0] + 
        ":obytes:AVERAGE"
      result << "DEF:obytes_max=" + Dir["#{prefix}*obytes.rrd"][0] + 
        ":obytes:MAX"
      result << "DEF:rbytes=" + Dir["#{prefix}*rbytes.rrd"][0] + 
        ":rbytes:AVERAGE"
      result << "CDEF:rbytes_neg=0,rbytes,-"
      result << "DEF:rbytes_max=" + Dir["#{prefix}*rbytes.rrd"][0] + 
        ":rbytes:MAX"
      result << "CDEF:rbytes_max_neg=0,rbytes_max,-"
      
      return result
    end
    
    # Returns the rrdgraph graphs for this graph.
    def graphs
      result = Array.new
      result << "AREA:obytes_max#777777"
      result << "AREA:rbytes_max_neg#777777"
      result << "AREA:obytes#770000:\"bytes out\""
      result << "AREA:rbytes_neg#007700:\"bytes in\""
      result << "HRULE:0#000000"
    end
  end

  # A graph showing cpu statistics.
  #
  # Shows percentage of cpu ticks used by kernel, user, wait and
  # idle. The graph is a stacked graph.
  class CpuGraph < GroupedGraph
    # Initializer
    def initialize
      super "cpu"
      add ServerstatGraphs::SimpleGraph.
        new("cpu_kernel", 
            "kstat_cpu_0_sys_cpu_ticks_kernel.rrd", 
            "cpu_ticks_kernel", 
            "AVERAGE", 
            "AREA", 
            "770000") 
      add ServerstatGraphs::SimpleGraph.
        new("cpu_user", 
            "kstat_cpu_0_sys_cpu_ticks_user.rrd", 
            "cpu_ticks_user", 
            "AVERAGE", 
            "AREA", 
            "007700", 
            true) 
      add ServerstatGraphs::SimpleGraph.
        new("cpu_wait", 
            "kstat_cpu_0_sys_cpu_ticks_wait.rrd", 
            "cpu_ticks_wait", 
            "AVERAGE", 
            "AREA", 
            "777700", 
            true) 
      add ServerstatGraphs::SimpleGraph.
        new("cpu_idle", 
            "kstat_cpu_0_sys_cpu_ticks_idle.rrd", 
            "cpu_ticks_idle", 
            "AVERAGE", 
            "AREA", 
            "777777",
            true) 
    end
  end

  # A graph showing different kinds of page faults.
  #
  # Shows minor (as) faults, copy-on-write faults, micro (hat) faults
  # and major faults.
  class FaultsGraph < GroupedGraph
    # Initializer
    def initialize
      super "faults"

      add ServerstatGraphs::SimpleGraph.
        new("as_fault", 
            "kstat_cpu_0_vm_as_fault.rrd", 
            "as_fault", 
            "AVERAGE", 
            "AREA", 
            "770000") 
      add ServerstatGraphs::SimpleGraph.
        new("cow_fault", 
            "kstat_cpu_0_vm_cow_fault.rrd", 
            "cow_fault", 
            "AVERAGE", 
            "AREA", 
            "777700",
            true) 
      add ServerstatGraphs::SimpleGraph.
        new("hat_fault", 
            "kstat_cpu_0_vm_hat_fault.rrd", 
            "hat_fault", 
            "AVERAGE", 
            "AREA", 
            "007700") 
      add ServerstatGraphs::SimpleGraph.
        new("maj_fault", 
            "kstat_cpu_0_vm_maj_fault.rrd", 
            "maj_fault", 
            "AVERAGE", 
            "AREA", 
            "007777",
            true) 
    end
  end

  # A graph showing the usage of a ZFS pool.
  #
  # The graph shows the usage of each volume in the pool and is
  # stacked.
  class ZpoolSpaceGraph < GroupedGraph
    # Initializer
    def initialize(pool)
      super pool + "_space"
      zpoolinfo = ZFS::ZpoolInfo.new(pool)
      zpoolinfo.getVolumes.each do |volume|
        defname = volume.gsub("/", "_")
        file = "zfs_" + pool + "_" +  defname + "_dataset.rrd"

        # All but the first subgraph are stacked.
        stacked = case graphs.length
                    when 0 then false
                    else true
                  end
        add ServerstatGraphs::SimpleGraph.
          new(defname,
              file,
              "dataset",
              "AVERAGE",
              "AREA",
              colorForVolume(volume),
              stacked)
      end

      # Finally add a subgraph for the available space on the pool.
      add ServerstatGraphs::SimpleGraph.
        new(pool + "_available",
            "zfs_" + pool + "_" + pool + "_available.rrd",
            "available",
            "AVERAGE",
            "AREA",
            "777777",
            true)
            
    end

    # Returns the color for a volume.
    #
    # [volume] the volume to compute the color for
    def colorForVolume(volume)
      color = "";
      index = 0;
      dirs = volume.split("/")

      # The HUE value is determined by the MD5 digest of the second
      # subvolume.
      h = if dirs.length > 1 then
            Digest::MD5.hexdigest(dirs[1])[0..1].hex
          else 
            0
          end

      # The saturation value is determined by the MD5 digest of the
      # third subvolume. It is greater or equal to 128.
      s = if dirs.length > 2 then
            (Digest::MD5.hexdigest(dirs[2])[0..1].hex % 128) + 128
          else
            255
          end

      # The intensity value is determined by the MD5 digest of the
      # last subvolume. It is greater or equal to 128.
      v = if dirs.length > 0 then
            (Digest::MD5.hexdigest(dirs.last)[0..1].hex % 128) + 128
          else
            255
          end

      # rrdgraph wants RGB
      rgbstrs = hsvtorgb(h, s, v).collect do |n|
        sprintf("%02x", n)
      end

      return rgbstrs.join
    end

    # Converts HSV to RGB.
    # 
    # [h] the HUE value (0..255)
    # [s] the saturation value (0..255)
    # [v] the intensity value (0..255)
    def hsvtorgb(h, s, v)
      # Based on the value of h decide which sector to use.
      # - hi is the number of the sector
      # - steps is the number of steps of h in the selected sector
      # - f is the number steps that h is into the selected sector
      case h
      when 0..43
        hi = 0
        steps = 43
        f = h
      when 44..85
        hi = 1
        steps = 85 - 44
        f = h - 44
      when 86..127
        hi = 2
        steps = 127 - 86
        f = h - 86
      when 128..170
        hi = 3
        steps = 213-128
        f = h - 128
      when 171..213
        hi = 4
        steps = 213 - 171
        f = h - 171
      when 214..255
        hi = 5
        steps = 255 - 214
        f = h - 214
      end


      # p is contant and determined by saturation
      p = (255 - s)

      # q goes down from 255 to p with f rising.
      q = p + s * (1 - f / steps)

      # t goes up from p to 255 with f rising.
      t = p + s * (1 - (steps - f) / steps)

      # scale by value
      p = p * v / 255
      q = q * v / 255
      t = t * v / 255

      # Construct return triple according to selected HUE sector.
      rgb = case hi
              when 0
              [v, t, p] # red to yellow
              when 1
              [q, v, p] # yellow to green
              when 2
              [p, v, t] # green to cyan
              when 3
              [p, q, v] # cyan to blue
              when 4
              [t, p, v] # blue to purple
              when 5
              [v, p, q] # purple to red
            end      
      
      return rgb
    end
  end

  # A simple HTTP server.
  #
  # The server serves a single request only in a given session.
  class HttpServer
    # Initializer
    #
    # [session] the session the request belongs to
    # [request] the request
    # [graphs] an Array containing the instances of Graph to be displayed
    def initialize(session, request, graphs)
      @session = session
      @request = request
      @graphs = graphs
    end

    # Output HTTP headers.
    #
    # [code] the return code (200 or 404)
    # [contentType] the content type of the contents served
    def headers(code, contentType)
      @session.print "HTTP/1.1 "

      case code
      when 200
        @session.print "200/OK\r\n"
        @session.print "Server: ServerstatGraphs\r\n"
        @session.print "Content-type: #{contentType}\r\n\r\n"
      when 404
        @session.print "404/Object Not Found\r\n"
        @session.print "Server: ServerstatGraphs\r\n"
      end
        
    end

    # Serves the graphs.
    #
    # For each graph a div is served containing a heading and a div
    # containing the img for the actual graph.
    #
    # The img has the following src:
    #
    # <tt>/graph/#{index}</tt>
    def serveGraphs
      headers(200, "text/html")
      @session.print "<html>"
      @session.print "<head>"
      @session.print "<title>ServerstatGraphs</title>"
      @session.print "</head>"
      @session.print "<body>"
      index = 0
      @graphs.each do |graph|
        @session.print "<div class=\"graphbox\">"
        @session.print "<h1>#{graph.name}</h1>"
        @session.print "<div class=\"graph\">"
        @session.print "<img src=\"/graph/#{index}\"/>"
        @session.print "</div>"
        @session.print "</div>"
        index += 1
      end
      @session.print "</body>"
      @session.print "</html>"
    end

    # Serves a single graph's PNG.
    #
    def serveGraph(graph)
      headers(200, "image/png")
      
      # Create the graph.
      system graph.rrdgraph
      
      # Open file and write contents to session.
      src = File.open(graph.filename)
      
      if (src != nil) then
        begin
          while (not src.eof?)
            buffer = src.read(256)
            @session.write(buffer)
          end
          src.close
        ensure
          src.close
        end
      end
    end

    # Serve the content.
    def serve()
      begin
        url = @request.split(" ")[1]

        case url
        when "/"
          serveGraphs
          
        when /\/graph\/.*/
          index = url.split("/")[2]
          graph = @graphs[index.to_i]
          serveGraph(graph)
        else
          headers(404, "")
        end
        
        @session.close
      end
    end
  end
end

def usage
  puts "Usage:"
  puts "#{$0} <command>*"
  puts ""
  puts "where <command> is"
  puts "defs                     returns the DEFs for all graphs"
  puts "graphs                   returns the rrdgraph definitions for all graphs"
  puts "rrdgraph                 returns the rrdgraph commands for all graphs"
  puts "create                   creates the graph's PNGs in $GRAPHPATH"
  puts "serve <ip>:<port>        starts server on <ip>:<port>"
end

Dir.chdir ENV['RRDPATH']

graphs = Array.new
graphs << ServerstatGraphs::CpuGraph.new
graphs << ServerstatGraphs::ZPoolGraph.new("tank", "tank")
graphs << ServerstatGraphs::NetworkGraph.new("e1000g", "e1000g");
graphs << ServerstatGraphs::FaultsGraph.new
graphs << ServerstatGraphs::ZpoolSpaceGraph.new("tank");

index = 0
while index < ARGV.count
  case ARGV[index]
  when "defs"
    graphs.each do |graph|
      p graph.defs
    end
  when "graphs"
    graphs.each do |graph|
      p graph.graphs
    end
  when "rrdgraph"
    graphs.each do |graph|
      print graph.rrdgraph + "\n"
    end
  when "create"
    graphs.each do |graph|
      system graph.rrdgraph
    end
  when "serve"
    if (index + 2 < ARGV.count) then
      server = TCPServer.new(ARGV[index + 1], ARGV[index + 2])
      loop do
        session = server.accept
        request = session.gets
        Thread.start(session, request, graphs) do |session, request, graphs|
          ServerstatGraphs::HttpServer.new(session, request, graphs).serve()
        end
      end
      
      index += 2
    end
  else
    puts "illegal command"
    usage
  end

  index += 1
end

if ARGV.length == 0 then
  usage
end
