#!/usr/bin/ruby
#
# serverstats 0.1
#
# Author::    Henning Brinkmann  (mailto:serverstats@henning-brinkmann.de)
# Copyright:: Copyright (c) 2011 Henning Brinkmann
# License::   GNU Public License
#
#
# serverstats is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# serverstats is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

require "./zfs"

# This module implements providers for server statistics.
module Serverstats

  # This is the base class for the providers.
  class Provider
    
    # Returns the filename for a given name.
    #
    # [name] the name to get the filename for
    def filename(name)
      return name
    end

    # Returns the dataset identifier for a given name.
    #
    # [name] the name to get the dataset identifier for 
    def ds(name)
      return name
    end

    # Fills the @data member with provider's data.
    def fillData
      # data format: [[name, value, kind]*]
      @data = Array.new
    end

    # Gets the providers data. The data is only gotten once and
    # afterwards returned from cache.
    def getData
      if @data == nil then
        self.fillData
      end
      
      return @data
    end

    # Determines the class of a given value.
    #
    # [str] the value to examine
    #
    # Return values:
    # GAUGE:: if <i>str</i> is a number containing a decimal point 
    #         (e.g. 123.45)
    # COUNTER:: if <i>str</i> is an integer number (e.g. 12345)
    # unknown:: otherwise
    def valueClass(str)
      case str
      when /[0-9]*[.][0-9]+/
        return "GAUGE"
      when /[0-9]+/
        return "COUNTER"
      else
        return "unknown"
      end
    end

    # Calls <tt>rrdcreate</tt> to create an RRD file for an entry in
    # the data returned by getData.
    #
    # [entry] entry from getdata to create the RRD file for
    #
    # The generated file will be named <tt>entry["filename"] +
    # ".rrd"</tt>. It will contain one dataset called
    # <tt>entry["ds"]</tt>. The value class will be
    # <tt>entry["class"]</tt>. The resolution of the dataset will be 2
    # minutes. There will be the following consolidation functions
    # used in the RRD file:
    #
    # - AVERAGE for every minute over 1 days
    # - AVERAGE for every 5 minutes over 1 week
    # - AVERAGE for every 10 minutes over 1  month
    # - AVERAGE for every hour over 1 year
    # - MAX for every minute over 1 days
    # - MAX for every 5 minutes over 1 week
    # - MAX for every 10 minutes over 1  month
    # - MAX for every hour over 1 year
    def rrdcreate(entry)
      filename = entry["filename"] + ".rrd"
      valueClass = entry["class"]
      if ! FileTest.exist?(filename) && valueClass != "unknown" then
        printf "Creating #{filename}\n"
        ds = entry["ds"]
        command = "rrdtool create #{filename} -s 60 \\\n"
        command += "        DS:#{ds}:#{valueClass}:120:U:U \\\n"
        command += "        RRA:AVERAGE:0.5:1:1440 \\\n"
        command += "        RRA:AVERAGE:0.5:5:2016 \\\n"
        command += "        RRA:AVERAGE:0.5:10:4320 \\\n"
        command += "        RRA:AVERAGE:0.5:60:8760 \\\n"
        command += "        RRA:MAX:0.5:1:1440 \\\n"
        command += "        RRA:MAX:0.5:5:2016 \\\n"
        command += "        RRA:MAX:0.5:10:4320 \\\n"
        command += "        RRA:MAX:0.5:60:8760 \\\n"
        system command
      end
    end

    # Calls <tt>rrdupdate</tt> to update the RRD file for an entry
    # with the according value.
    #
    # [entry] entry for which to run the update
    #
    # If <tt>entry["class"]</tt> is <tt>unknown</tt> nothing will be
    # done. Otherwise, the command executed will be
    #
    # <tt>rrdtool update #{entry["filename"]}.rrd N:#{entry["value"]}</tt>
    def rrdupdate(entry)
      filename = entry["filename"] + ".rrd"
      valueClass = entry["class"]
      if valueClass != "unknown" then
        value = entry["value"]
        command = "rrdtool update #{filename} N:#{value}"
        system command
      end
    end

    # Creates the RRD file for each entry in the data returned by
    # getData.
    #
    # This method uses rrdcreate.
    def create
      getData.each do |entry|
        rrdcreate entry
      end
    end

    # Updates the RRD file for each entry in the data returned by
    # getData.
    #
    # This method uses rrdupate.
    def update
      getData.each do |entry|
        rrdupdate entry
      end
    end
  end

  # A Provider for data retrieved from <tt>kstat</tt>
  class KStatProvider < Provider

    # Initializer
    #
    # [mod] the kstat module to get data for
    # [instance] the kstat instance to get data for
    # [name] the kstat name to get data for
    def initialize(mod, instance, name)
      @mod = mod
      @instance = instance
      @name = name
    end

    # Return the filename for a given kstat name.
    #
    # The name is generated by adding the prefix "kstat_" and
    # replacing all ":" by "_".
    #
    # [name] the kstat name to generate the filename for
    def filename(name)
      return "kstat_" + name.gsub(":", "_")
    end

    # Returns the dataset identifier for a given kstat name.
    #
    # The dataset name is the last filed when splitting at ":".
    #
    # [name] the kstat name to generate the dataset identifier for
    def ds(name)
      return name.split(":").last
    end

    # Returns the command to be run for retrieving the data.
    #
    # The command to be run is
    #
    # <tt>kstat -p -m #{@mod} -i #{@instance} -n #{@name}</tt>
    def command
      return "kstat -p -m #{@mod} -i #{@instance} -n #{@name}"
    end

    # Fills the data.
    def fillData
      f = IO.popen(command)
      @data = Array.new
      f.readlines.each do |line|
        # The lines returned by kstat have two fields delimited by
        # tabs. The first field is the name of the data item. The
        # second is the value.
        #
        # Sample line from kstat:
        # cpu:0:sys:syscall       139315104
        parts = line.chop.split("\t");
        if parts[0] != nil && parts[1] != nil then
          entry = Hash.new
          entry["filename"] = filename(parts[0])
          entry["ds"] = ds(parts[0])
          entry["value"] = parts[1]
          entry["class"] = valueClass(parts[1])
          @data << entry
        end
      end
    end
  end

  # This class implements a provider that retrieves the kstat data for
  # the disks in a ZFS pool.
  class ZpoolProvider < KStatProvider

    # Initializer
    # 
    # [pool] the ZFS pool to get data for
    def initialize(pool)
      @info = ZFS::ZpoolInfo.new(pool)
      @pool = pool
    end

    # The filename for a specific data set.
    #
    # The filename is generated by replacing all ":" by "_" and adding
    # the prefix "zpool_".
    #
    # [name] the kstat name to generate the filename for
    def filename(name)
      return "zpool_" + @pool + "_" + name.gsub(":", "_")
    end

    # For each disk sd in the ZFS pool the following command is called_
    #
    # <tt>kstat -p -n #{sd}</tt>
    #
    # The command is generated by joining the several commands using
    # "&&".
    def command
      tmp = @info.getSDs.collect do |sd|
        "kstat -p -n #{sd}"
      end

      return tmp.join("&&")
    end
  end

  # This class uses <tt>/usr/sbin/zfs</tt> to provide data about the
  # space consumption of the volumes of a ZFS filesystem and its
  # descendants.
  class ZFSProvider < Provider

    # Initializer
    #
    # [fs] the ZFS filesystem to provide data for
    def initialize(fs)
      @fs = fs
    end

    # Returns filename for a ZFS filesystem.
    #
    # [name] the ZFS filesystem to generate the filename for
    #
    # The filename is generated by replacing all "/" by "_" and adding
    # the prefix <tt>zfs_#{fs}_</tt>.
    def filename(name)
      return "zfs_" + @fs + "_" + name.gsub("/", "_")
    end

    # Returns the dataset identifier for a ZFS filesystem.
    # 
    # [name] the ZFS filesystem to generate the dataset name for
    #
    # The dataset identifier is generated by splitting the filesystem
    # name at "/" and taking the last field.
    def ds(name)
      return name.split("/").last
    end

    # Fills the data.
    #
    # For each descendant filesystem of @fs there will be an entries like
    # this:
    # 
    # <tt>{"filename"=>"zfs_tank_tank_www_children", "ds"=>"children",
    # "value"=>"0", "class"=>"ABSOLUTE"}</tt>
    #
    # The example represents the ZFS filesystem <tt>tank/www</tt> when
    # <tt>@fs=="tank"</tt>.
    def fillData
      # Get the descendant filesystems.
      command = "/usr/sbin/zfs list -Hr -o name #{@fs}"
      f = IO.popen(command)
      filesystems = f.readlines.collect do |line|
        line.chop
      end

      # For each filesystem generate the entries for that filesystem.
      @data = Array.new
      filesystems.each do |fs|
        # Get the space data from the current filesystem.
        #
        # E.g:
        # tank    name    tank    -
        # tank    available       787890899144    -
        # tank    used    2645708581888   -
        # tank    usedbysnapshots 0       -
        # tank    usedbydataset   48128   -
        # tank    usedbyrefreservation    0       -
        # tank    usedbychildren  2645708533760   -
        command = "/usr/sbin/zfs get -Hp space #{fs}"
        f = IO.popen(command)
        lines = f.readlines
        lines.each do |line|
          parts = line.split("\t")
          if parts[1] != "name" then
            # We will ignore the line defining the name as the name is
            # specified as the first field anyway. The name to genrate
            # the filename and the dataset identifier from will by
            # concatenated using "/" from the filesystem name and the
            # property name. Every substring "usedby" will be removed.
            name = parts[0] + "/" + parts[1].gsub("usedby", "")
            entry = Hash.new
            entry["filename"] = filename(name)
            entry["ds"] = ds(name)
            entry["value"] = parts[2]
            entry["class"] = "ABSOLUTE"
            @data << entry 
          end
        end
      end
    end
  end
end

# Output usage information.
def usage
  puts "#{$0} creates and updates RRD databases with server statistics data."
  puts "Usage:"
  puts "#{$0} <command>+"
  puts ""
  puts "where [command] is one of the following"
  puts ""
  puts "getdata       Get the data and print it out."
  puts "unknowndata   Output data with class \"unknown\""
  puts "create        Create RRD databases if they are not already existing."
  puts "update        Feed current data to RRD databases"
end

# Change to the input directory for RRD databases.
Dir.chdir ENV['RRDPATH']

# Create providers.
providers = Array.new
providers << Serverstats::KStatProvider.new("cpu", "0", "sys")
providers << Serverstats::KStatProvider.new("cpu", "0", "vm")
providers << Serverstats::KStatProvider.new("e1000g", "0", "mac")
providers << Serverstats::ZpoolProvider.new("tank")
providers << Serverstats::ZFSProvider.new("tank")

# Handle arguments.
index = 0
while index < ARGV.count
  case ARGV[index]
  when "getdata"
    providers.each do |provider|
      p provider.getData
    end
  when "unknowndata"
    providers.each do |provider|
      provider.getData.each do |entry|
        if entry["class"] == "unknown" then
          p entry
        end
      end
    end
  when "create"
    providers.each do |provider|
      provider.create
    end
  when "update"
    providers.each do |provider|
      provider.update
    end
  when "usage"
    usage
  else
    puts "Illegal command."
    puts "Try #{$0} usage."
  end

  index += 1
end

if ARGV.count == 0 then
  usage
end
