#
# serverstats 0.1
#
# Author::    Henning Brinkmann  (mailto:serverstats@henning-brinkmann.de)
# Copyright:: Copyright (c) 2011 Henning Brinkmann
# License::   GNU Public License
#
#
# serverstats is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# serverstats is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <http://www.gnu.org/licenses/>.

# This module provides helper classes for ZFS pools.
module ZFS
  # This class provides information about a ZFS pool.
  class ZpoolInfo
    # Initializer
    #
    # [pool] the pool
    def initialize(pool)
      @pool = pool
    end

    # Returns mapping from disk identifiers in "ctd" format to "sd"
    # format.
    #
    # The returned Hash uses the "ctd"s as keys and "sd"s as
    # values. I.e. <tt>getCTD2sd["c9t0d0"]</tt> might give "sd0" as an
    # result.
    #
    # The shell command <tt>iostat</tt> is used to get the mapping.
    def getCTD2sd
      if @CTD2sd == nil then
        # The "sd"s will be in sds.
        # The "ctd"s till be in ctds.

        # Get iostat containing the "sd"s.
        f = IO.popen("iostat -xr")
        lines = f.readlines

        # The data is in the lines following the header.
        sds = lines[2..lines.count - 1].collect do |line|
          # The "sd"s are in the first field.
          line.split(",")[0]
        end

        # Get iostat containing the "ctd"s.
        f = IO.popen("iostat -xrn")
        lines = f.readlines

        # The data is in the line following the header.
        ctds = lines[2..lines.count - 1].collect do |line|
          # The "ctd"s are in the last, i.e. tenth, field.
          line.chop.split(",")[10]
        end

        # Create the hash. The equivalent "ctd"s and "sd"s are
        # correlated by the position in the arrays containing "cts"s
        # and "sd"s (ctds and sds).
        @CTD2sd = Hash.new
        0.upto(sds.count-1) do |i| 
          @CTD2sd[ctds[i]] = sds[i]
        end
      end
      
      return @CTD2sd
    end

    # Returns an Array containing the "sd"s in the pool.
    def getSDs
      if @sds == nil then
        # Get the verbose iostat of the pool which contains the names
        # of all "ctd"s in the pool.
        f = IO.popen("/usr/sbin/zpool iostat -v #{@pool}")
        lines = f.readlines
        @sds = Array.new
        
        # Interesting data is in the lines following the first three
        # lines and excluding the trailing two lines.
        lines[3..lines.count-3].each do |line|
          # Remove formatting spaces and take the first field. That
          # might contain the "ctd".
          ctd = line.squeeze(" ").strip.split(" ")[0]

          # Map the "ctd" candidate to the according "sd". If the
          # mapping succeeds add the "sd" to the result.
          sd = getCTD2sd[ctd]
          if sd != nil then
            @sds << sd
          end
        end
      end

      return @sds
    end

    # Returns an Array containing the volumes in the pool.
    def getVolumes
      if @volumes == nil then
        # Get the names of the volumes in the pool via zfs list and
        # put them into the result.
        f = IO.popen("/usr/sbin/zfs list -Hro name #{@pool}")
        lines = f.readlines
        @volumes = Array.new
        lines.each do |line|
          @volumes << line.chomp
        end
      end

      return @volumes
    end

  end

end
